#include "opengl.hh"

#include <glow/common/log.hh>

#include "nanovg/nanovg.hh"
#include "nanovg/nanovg_gl.hh"

#include "../image2D.hh"

using namespace glow;
using namespace glow::vector;


static int vg_cap(line_cap c)
{
    switch (c)
    {
    case line_cap::butt:
        return nvg::NVG_BUTT;
    case line_cap::round:
        return nvg::NVG_ROUND;
    case line_cap::square:
        return nvg::NVG_SQUARE;
    default:
        return 0;
    }
};
static int vg_join(line_join j)
{
    switch (j)
    {
    case line_join::bevel:
        return nvg::NVG_BEVEL;
    case line_join::miter:
        return nvg::NVG_MITER;
    case line_join::round:
        return nvg::NVG_ROUND;
    default:
        return 0;
    }
};

OGLRenderer::~OGLRenderer()
{
    if (vg)
        nvg::nvgDeleteGL3(vg);
}

void OGLRenderer::render(image2D const& img, int w, int h, float devicePixelRatio) const
{
    using namespace nvg;

    lazy_init();

    struct nvg_renderer
    {
        NVGcontext* vg;

        void begin_path() { nvgBeginPath(vg); }
        void close_path() { nvgClosePath(vg); }

        void move_to(float x, float y) { nvgMoveTo(vg, x, y); }
        void line_to(float x, float y) { nvgLineTo(vg, x, y); }
        void bezier_to(float c1x, float c1y, float c2x, float c2y, float x, float y) { nvgBezierTo(vg, c1x, c1y, c2x, c2y, x, y); }
        void quad_to(float cx, float cy, float x, float y) { nvgQuadTo(vg, cx, cy, x, y); }
        void arc_to(float x1, float y1, float x2, float y2, float radius) { nvgArcTo(vg, x1, y1, x2, y2, radius); }

        // TODO: arc, rounded_rect_varying, ellipse
        void rect(float x, float y, float w, float h) { nvgRect(vg, x, y, w, h); }
        void rounded_rect(float x, float y, float w, float h, float r) { nvgRoundedRect(vg, x, y, w, h, r); }
        void circle(float cx, float cy, float r) { nvgCircle(vg, cx, cy, r); }

        void stroke() { nvgStroke(vg); }
        void fill() { nvgFill(vg); }

        void set_stroke_width(float s) { nvgStrokeWidth(vg, s); }
        void set_stroke_color(float r, float g, float b, float a) { nvgStrokeColor(vg, {{{r, g, b, a}}}); }
        void set_fill_color(float r, float g, float b, float a) { nvgFillColor(vg, {{{r, g, b, a}}}); }
        void set_line_cap(line_cap cap) { nvgLineCap(vg, vg_cap(cap)); }
        void set_line_join(line_join join) { nvgLineJoin(vg, vg_join(join)); }
        void set_miter_limit(float limit) { nvgMiterLimit(vg, limit); }

        void clip_rect(float x, float y, float w, float h) { nvgScissor(vg, x, y, w, h); }
        void reset_clip() { nvgResetScissor(vg); }
    };

    nvgBeginFrame(vg, w, h, devicePixelRatio);
    visit(img, nvg_renderer{vg});
    nvgEndFrame(vg);
}

void OGLRenderer::loadFont(const std::string& name, const std::string& file)
{
    lazy_init();

    auto f = nvg::nvgCreateFont(vg, name.c_str(), file.c_str());
    if (f == -1)
        glow::error() << "could not create font '" << name << "' from file '" << file << "'";
}

void OGLRenderer::lazy_init() const
{
    if (!vg)
        vg = nvg::nvgCreateGL3(nvg::NVG_ANTIALIAS | nvg::NVG_STENCIL_STROKES);
}
