#pragma once

#include "image2D.hh"

#include <typed-geometry/tg.hh>

namespace glow
{
namespace vector
{
// TODO: gradients / paints
struct stroke2D
{
    tg::color4 color;
    float width = 1.0f;

    stroke2D() = default;
    stroke2D(tg::color3 const& c) : color(c, 1.0f), width(1.0f) {}
    stroke2D(tg::color4 const& c, float w = 1.0f) : color(c), width(w) {}
    stroke2D(float r, float g, float b, float a = 1.0f) : color(r, g, b, a), width(1.0f) {}
};
struct fill2D
{
    tg::color4 color;

    fill2D() = default;
    fill2D(tg::color4 const& c) : color(c) {}
    fill2D(tg::color3 const& c) : color(c) {}
    fill2D(float r, float g, float b, float a = 1.0f) : color(r, g, b, a) {}
};

/// High-level API for writing to images
/// See Vector2DSample in glow-samples
///
/// Usage:
///   auto g = graphics(img);
///   g.draw(<tg-obj>, <stroke>);
///   g.fill(<tg-obj>, <fill>);
///
/// Example:
///   g.draw(tg::pos3{...}, {1, 0, 0}); // 1px red dot
///   g.draw(tg::segment3{...}, {tg::color3::blue, 5.0f}); // 5px blue segment
///
/// TODO:
///   - infinite types like ray and line
///   - vectors
///   - think about customization points
struct graphics2D
{
    explicit graphics2D(image2D& img) : img(img) {}

    // drawing
public:
    void draw(tg::pos2 const& p, stroke2D const& stroke)
    {
        img.begin_path();
        img.rect({p - .5f, p + .5f});
        perform_draw(stroke);
    }

    void draw(tg::segment2 const& s, stroke2D const& stroke)
    {
        img.begin_path();
        img.move_to(s.pos0);
        img.line_to(s.pos1);
        perform_draw(stroke);
    }

    void draw(tg::aabb2 const& r, stroke2D const& stroke)
    {
        img.begin_path();
        img.rect(r);
        perform_draw(stroke);
    }

    void draw(tg::triangle2 const& t, stroke2D const& stroke)
    {
        img.begin_path();
        img.move_to(t.pos0);
        img.line_to(t.pos1);
        img.line_to(t.pos2);
        img.line_to(t.pos0);
        perform_draw(stroke);
    }

    void draw(tg::circle2 const& c, stroke2D const& stroke)
    {
        img.begin_path();
        img.circle(c.center, c.radius);
        perform_draw(stroke);
    }

    void draw(tg::sphere2 const& s, stroke2D const& stroke)
    {
        img.begin_path();
        img.circle(s.center, s.radius);
        perform_draw(stroke);
    }

    // filling
public:
    void fill(tg::aabb2 const& r, fill2D const& fill)
    {
        img.begin_path();
        img.rect(r);
        perform_fill(fill);
    }

    void fill(tg::triangle2 const& t, fill2D const& fill)
    {
        img.begin_path();
        img.move_to(t.pos0);
        img.line_to(t.pos1);
        img.line_to(t.pos2);
        img.line_to(t.pos0);
        perform_fill(fill);
    }

    void fill(tg::circle2 const& c, fill2D const& fill)
    {
        img.begin_path();
        img.circle(c.center, c.radius);
        perform_fill(fill);
    }

    void fill(tg::sphere2 const& s, fill2D const& fill)
    {
        img.begin_path();
        img.circle(s.center, s.radius);
        perform_fill(fill);
    }

public:
    image2D_low_level_api& low_level_api() { return img; }

private:
    void perform_draw(stroke2D const& s)
    {
        img.set_stroke_color(s.color);
        img.set_stroke_width(s.width);
        img.stroke();
    }
    void perform_fill(fill2D const& f)
    {
        img.set_fill_color(f.color);
        img.fill();
    }

    image2D_low_level_api img;
};

inline graphics2D graphics(image2D& img) { return graphics2D(img); }
}
}
