#include "SubView.hh"

#include <glm/ext.hpp>

#include <glow/common/scoped_gl.hh>
#include <glow/data/SurfaceData.hh>
#include <glow/data/TextureData.hh>
#include <glow/math/transform.hh>
#include <glow/objects/Framebuffer.hh>
#include <glow/objects/TextureCubeMap.hh>
#include <glow/objects/UniformBuffer.hh>
#include <glow/std140.hh>

#include <glow-extras/glfw/LambdaApp.hh>
#include <glow-extras/glfw/LambdaPipeline.hh>
#include <glow-extras/material/IBL.hh>
#include <glow-extras/pipeline/RenderPipeline.hh>
#include <glow-extras/pipeline/RenderScene.hh>
#include <glow-extras/pipeline/stages/StageCamera.hh>

#include <GLFW/glfw3.h>

#include "Scene.hh"
#include "renderables/Renderable.hh"
#include "view.hh"

#include "ViewerApp.hh"

using namespace glow;
using namespace glow::viewer;

void SubView::initRendering()
{
    // set up camera
    assert(mCamera && "no camera set");

    // set up scene
    auto aabb = mScene->computeAabb();
    if (aabb.isEmpty())
        glow::warning() << "Scene has empty AABB";
    else
    {
        // camera position
        auto size = aabb.size();
        auto center = aabb.center();
        auto dir = normalize(normalize(aabb.max - center) + glm::vec3(0, 1, 1));
        auto diameter = length(size);

        float maxdis = glm::max(size.x, glm::max(size.y, size.z));
        mScene->setScale(maxdis / 20);

        auto const verticalFovRad = glm::radians(mCamera->getVerticalFieldOfView());
        auto const horizontalFovRad = glm::radians(mCamera->getHorizontalFieldOfView());
        float targetDistance = diameter / (2 * glm::tan(verticalFovRad / 2.f));
        targetDistance = std::max(targetDistance, static_cast<float>(diameter / (2 * tan(horizontalFovRad / 2.f))));

        mCamera->setLookAt(center + dir * targetDistance * 2.f, center);
        mCamera->setNearPlane(diameter * 1e-4f);
        mCamera->setFarPlane(diameter * 10);
        mCamera->handle.snap();

        mPipelineScene->edgeOutline.depthThreshold = mScene->getScale() * 2.5f;
    }

    // TODO: subviews
}

void SubView::performLayout(glm::ivec2 start, glm::ivec2 size)
{
    mViewportStart = start;
    mViewportSize = size;

    if (mLayout)
        mLayout->performLayout(*this);
}

void SubView::show()
{
    try
    {
        ViewerApp app(*this);
        app.run();
    }
    catch (ViewerApp::TerminateException&)
    {
        std::exit(0);
    }
}

void SubView::addSubView(const SharedSubView& view) { mSubViews.push_back(view); }

SharedTextureRectangle SubView::render(int w, int h, const std::string& fileName)
{
    glow::error() << "Implement me";
    /*
    init();

    mPipelineApp->startHeadless();

    // Camera setup
    {
        if (w <= 0)
            w = 3840;

        if (h <= 0)
            h = 2160;

        mPipelineApp->getCamera()->setViewportSize(w, h);
        adaptToScene();
    }

    // Warmup / TAA Accumulation
    for (int i = 0; i < 3; ++i)
        mPipelineApp->getPipeline()->render(*mPipelineApp->getPipelineScene(), *mPipelineApp);

    auto outputTexture = glow::TextureRectangle::create(static_cast<int>(mPipelineApp->getCamera()->getViewportWidth()),
                                                        static_cast<int>(mPipelineApp->getCamera()->getViewportHeight()), GL_RGB);

    {
        auto outputFbo = glow::Framebuffer::create("fOut", outputTexture);
        auto const fb = outputFbo->bind();
        mPipelineApp->getPipeline()->render(*mPipelineApp->getPipelineScene(), *mPipelineApp);
    }

    if (fileName != "")
        outputTexture->bind().writeToFile(fileName);

    return outputTexture;*/
    return nullptr;
}

SubView::SubView()
{
    Viewer::GlobalInit(); // lazy init

    mScene = Scene::create();

    // create pipeline scene and camera
    mPipelineScene = pipeline::RenderScene::create();
    mPipelineCamera = pipeline::StageCamera::create();

    // adjust pipeline settings
    {
        auto& s = *mPipelineScene;

        // Enable edge outline
        s.edgeOutline.enabled = true;

        // Disable atmospheric scattering
        s.atmoScatter.intensity = 0.f;
        s.sun.color = {1, 1, 1};
        s.sun.intensity = 1.f;

        s.ao.largeScaleIntensity = .47f;
        s.ao.smallScaleIntensity = 1.f;

        // Neutral postprocessing
        s.colorLut = nullptr;
        s.tonemappingEnabled = false;
        s.exposure = 1.0;
        s.contrast = 1.0f;
        s.sharpenStrength = 0.f;

        // Tone down bloom
        s.bloomThreshold = 3.f;
        s.bloomIntensity = 1.15f;
    }
}

SharedSubView SubView::create() { return std::make_shared<SubView>(); }
