#include "ViewerApp.hh"

#include <glow/common/scoped_gl.hh>
#include <glow/objects/Framebuffer.hh>
#include <glow/objects/Program.hh>
#include <glow/objects/Shader.hh>
#include <glow/objects/TextureCubeMap.hh>
#include <glow/objects/UniformBuffer.hh>
#include <glow/std140.hh>

#include <GLFW/glfw3.h>

#include <glow-extras/debugging/imgui-util.hh>
#include <glow-extras/geometry/Quad.hh>
#include <glow-extras/material/IBL.hh>
#include <glow-extras/pipeline/RenderPipeline.hh>
#include <glow-extras/pipeline/RenderScene.hh>

#include <imgui/imgui.h>

#include "SubView.hh"
#include "view.hh"

using namespace glow;
using namespace glow::viewer;

static glow::SharedTextureCubeMap sDefaultEnvMap;

namespace
{
struct ShaderGlobalUbo
{
    glow::std140mat4x4 view;
    glow::std140mat4x4 proj;
    glow::std140mat4x4 invView;
    glow::std140mat4x4 invProj;
    glow::std140mat4x4 cleanVp;
    glow::std140mat4x4 prevCleanVp;

    glow::std140vec3 camPos;
    glow::std140vec3 sunDirection;
    glow::std140vec3 sunColor;
};

template <class View, class F>
void visitLeafViews(View&& view, F&& renderF)
{
    if (view.isContainer())
    {
        for (auto const& v : view.getSubViews())
            visitLeafViews(*v, renderF);
    }
    else
        renderF(view);
}
}

ViewerApp::ViewerApp(SubView& view) : GlfwApp(Gui::ImGui), mRootView(view) {}

void ViewerApp::init()
{
    setDumpTimingsOnShutdown(false);
    setUsePipeline(false);
    setUsePipelineConfigGui(false);
    setTitle("Viewer");

    GlfwApp::init();

    mPipeline = pipeline::RenderPipeline::createDefaultPipeline();

    mUboGlobals = glow::UniformBuffer::create();
    mUboGlobals->bind().setData(ShaderGlobalUbo{});

    auto const& skybox = Viewer::getSkybox();
    if (skybox)
        mEnvironmentMap = material::IBL::createEnvMapGGX(skybox);
    else
        mEnvironmentMap = material::IBL::getDefaultCubeMap();

    visitLeafViews(mRootView, [&](SubView& v) {
        // shared cam for now
        if (!v.camera())
            v.camera(getCamera());

        // init view (adapt to scene etc)
        v.initRendering();

        // initialize renderables
        // TODO: prevent multi init
        for (auto const& r : v.scene()->getRenderables())
            r->init();

        // adjust camera speed
        // AFTER initRendering
        getWASDController().setCameraSpeed(v.scene()->getScale() * 20.f);
    });

    // init context
    mContext.app = this;
    mContext.root_view = &mRootView;

    debugging::applyGlowImguiTheme();
}

void ViewerApp::update(float)
{
    // empty for now
}

void ViewerApp::render(float)
{
#ifdef GLOW_EXTRAS_OPENGL_DEBUG_GROUPS
    GLOW_SCOPED(debugGroup, "Render Viewer");
#endif

    // apply layout
    mRootView.performLayout({0, 0}, getWindowSize());

    // clear black
    glClearColor(0, 0, 0, 1);
    glClear(GL_COLOR_BUFFER_BIT);

    // render each view
    visitLeafViews(mRootView, [&](SubView const& v) {
        // start context
        mContext.view = &v;
        mContext.scene = v.scene().get();

        // update pipeline scene
        {
            auto& s = *v.mPipelineScene;
            s.ao.radius = v.scene()->getScale();
            s.backgroundColor = v.scene()->backgroundColor();
        }

        // set up pipeline cam
        auto vs = v.getViewportSize();
        // glow::info() << to_string(vs);
        v.camera()->setFoV(60.f);
        v.camera()->setViewportSize(vs.x, vs.y);
        v.mPipelineCamera->onNewFrame(*v.camera());

#ifdef GLOW_EXTRAS_OPENGL_DEBUG_GROUPS
        GLOW_SCOPED(debugGroup, "SubView Render");
#endif

        // set viewport
        auto start = v.getViewportStart();
        auto const& size = v.getViewportSize();
        start.y = getWindowHeight() - start.y - size.y - 1;
        mPipeline->restrictOutputViewport(start, size);

        // render
        mPipeline->render(*v.mPipelineCamera, *v.mPipelineScene, *this);
    });
}

void ViewerApp::onPerformShadowPass(pipeline::RenderContext const& ctx, UsedProgram& shader)
{
    // TODO
}

void ViewerApp::onRenderDepthPrePass(pipeline::RenderContext const& ctx)
{
    mContext.setRenderContext(&ctx);

    updateUniformBuffer(ctx);

    for (auto const& r : mContext.view->scene()->getRenderables())
        r->renderDepthPre(mContext);
}

void ViewerApp::onRenderOpaquePass(pipeline::RenderContext const& ctx)
{
    mContext.setRenderContext(&ctx);

    updateUniformBuffer(ctx);

    for (auto const& r : mContext.view->scene()->getRenderables())
        r->renderForward(mContext);
}

void ViewerApp::onRenderTransparentPass(pipeline::RenderContext const& ctx)
{
    mContext.setRenderContext(&ctx);

    updateUniformBuffer(ctx);

    for (auto const& r : mContext.view->scene()->getRenderables())
        r->renderTransparent(mContext);
}

bool ViewerApp::onMouseButton(double x, double y, int button, int action, int mods, int clickCount)
{
    // TODO: properly per subview

    if (button == GLFW_MOUSE_BUTTON_LEFT && action != GLFW_RELEASE && clickCount > 1 && mods == 0)
    {
        if (false)
        {
            auto const recvPos = getPipeline()->queryPosition3D(*getPipelineCamera(), glm::uvec2(static_cast<unsigned>(x), static_cast<unsigned>(y)));
            if (recvPos)
            {
                getCamera()->handle.translateFocus(recvPos.value());
                return true;
            }
        }
        else
            glow::warning() << "not implemented";
    }

    if (GlfwApp::onMouseButton(x, y, button, action, mods, clickCount))
        return true;

    return false;
}

bool ViewerApp::onKey(int key, int scancode, int action, int mods)
{
    if (GlfwApp::onKey(key, scancode, action, mods))
        return true;

    if (action == GLFW_PRESS)
    {
        if (key == GLFW_KEY_ESCAPE)
        {
            if (input().isKeyDown(GLFW_KEY_LEFT_SHIFT))
            {
                // Terminate on Shift + ESC
                throw TerminateException();
            }
            else
            {
                // Close on ESC
                requestClose();
            }
            return true;
        }
    }

    return false;
}

void ViewerApp::onResize(int w, int h)
{
    GlfwApp::onResize(w, h);
}

void ViewerApp::onGui()
{
    if (ImGui::BeginMainMenuBar())
    {
        ImGui::PushID("glow_viewer_menu");

        ImGui::MenuItem("Viewer", nullptr, false, false);

        if (ImGui::BeginMenu("Overlays"))
        {
            if (ImGui::MenuItem("Profiling", "F9"))
                toggleProfilingOverlay();

            if (ImGui::MenuItem("OpenGL Log", "F10"))
                toggleDebugOverlay();


            ImGui::EndMenu();
        }

        ImGui::EndMainMenuBar();

        ImGui::PopID();
    }

    // perform imgui for each renderable in leaf subviews
    visitLeafViews(mRootView, [&](SubView const& v) {
        for (auto const& r : v.scene()->getRenderables())
            r->performImgui();
    });
}

void ViewerApp::prepareShader(const SharedProgram& shader, UsedProgram& usedShader) const
{
    if (mRegisteredForwardShaders.find(shader.get()) == mRegisteredForwardShaders.end())
    {
        shader->setUniformBuffer("uViewerGlobalsUBO", mUboGlobals);
        mRegisteredForwardShaders.insert(shader.get());
    }

    material::IBL::prepareShaderGGX(usedShader, mEnvironmentMap);
}

void ViewerApp::updateUniformBuffer(pipeline::RenderContext const& ctx)
{
    mUboGlobals->bind().setData(
        ShaderGlobalUbo{ctx.camData.view, ctx.camData.proj, ctx.camData.viewInverse, ctx.camData.projInverse, ctx.camData.cleanVp,
                        ctx.camData.prevCleanVp, ctx.camData.camPos, ctx.scene.sun.direction, ctx.scene.sun.color * ctx.scene.sun.intensity},
        GL_STREAM_DRAW);
}
