#pragma once

#include <initializer_list>
#include <sstream>
#include <string>

#include <glm/ext.hpp>
#include <glm/glm.hpp>

namespace glow
{
namespace viewer
{
struct aabb
{
    // member
public:
    glm::vec3 min = glm::vec3(+std::numeric_limits<float>::max());
    glm::vec3 max = glm::vec3(-std::numeric_limits<float>::max());

    // properties
public:
    glm::vec3 size() const;
    glm::vec3 center() const;
    float radius() const; ///< Radius of the bounding sphere (positioned at the center)
    float volume() const;
    bool isEmpty() const; ///< NOT the same as zero volume! 2D aabb is not empty.

    // mutating methods
public:
    /// includes a given point
    void add(glm::vec3 const& p);
    /// includes a given aabb
    void add(aabb const& rhs);

    // non-mutating
public:
    /// calculates a transformed aabb
    aabb transformed(glm::mat4 const& transform) const;

    // ctor
public:
    aabb() = default;
    aabb(glm::vec3 min, glm::vec3 max) : min(min), max(max) {}

    static aabb empty() { return {}; }
};

/// ======== IMPLEMENTATION ========

inline glm::vec3 aabb::size() const
{
    return max - min;
}

inline glm::vec3 aabb::center() const
{
    return (max + min) * .5f;
}

inline float aabb::radius() const
{
    return glm::length(size()) * .5f;
}

inline float aabb::volume() const
{
    if (isEmpty())
        return 0;

    auto s = size();
    return s.x * s.y * s.z;
}

inline bool aabb::isEmpty() const
{
    return max.x < min.x || max.y < min.y || max.z < min.z;
}

inline void aabb::add(const glm::vec3& p)
{
    min = glm::min(min, p);
    max = glm::max(max, p);
}

inline void aabb::add(const aabb& rhs)
{
    add(rhs.min);
    add(rhs.max);
}

inline aabb aabb::transformed(const glm::mat4& transform) const
{
    aabb r;
    for (int dz : {0, 1})
        for (int dy : {0, 1})
            for (int dx : {0, 1})
            {
                auto p = min + (max - min) * glm::vec3(dx, dy, dz);
                p = glm::vec3(transform * glm::vec4(p, 1.0f));
                r.add(p);
            }
    return r;
}

inline std::string to_string(aabb const& aabb)
{
    std::stringstream ss;
    ss << "aabb[" << to_string(aabb.min) << ", " << to_string(aabb.max) << "]";
    return ss.str();
}
}
}
