#include "LineRenderable.hh"

#include <glow-extras/material/IBL.hh>

#include "../SubView.hh"
#include "../ViewRenderContext.hh"
#include "../detail/MeshShaderBuilder.hh"

using namespace glow;
using namespace glow::viewer;

aabb LineRenderable::computeAabb()
{
    // TODO: enlarge by line width
    return getMeshAABB().transformed(transform());
}

void LineRenderable::renderShadow(ViewRenderContext const& ctx, UsedProgram& shader)
{
    // nothing for now
}

void LineRenderable::renderDepthPre(ViewRenderContext const& ctx)
{
    // nothing for now
}

void LineRenderable::renderForward(ViewRenderContext const& ctx)
{
    auto shader = ctx.use(mForwardShader);

    shader.setUniform("uModel", transform());
    shader.setUniform("uInvModel", inverse(transform()));
    shader.setUniform("uRoundCaps", mRoundCaps);
    shader.setUniform("uNoCaps", mNoCaps);
    shader.setUniform("uExtrapolate", mExtrapolate);
    shader.setUniform("uWorldSpaceSize", mWorldSpaceSize);
    shader.setUniform("uCameraFacing", mCameraFacing);
    shader.setUniform("uScreenSize", glm::vec2(ctx.render_camera_data->resolution));

    if (getColorMapping())
        getColorMapping()->prepareShader(shader);
    if (getTexturing())
        getTexturing()->prepareShader(shader);
    for (auto const& a : getAttributes())
        a->prepareShader(shader);

    mVertexArray->bind().draw();
}

void LineRenderable::renderTransparent(ViewRenderContext const& ctx)
{
    // nothing for now
}

SharedLineRenderable LineRenderable::create(const builder::LineBuilder& builder)
{
    auto r = std::make_shared<LineRenderable>();
    r->initFromBuilder(builder);
    return r;
}

void LineRenderable::initFromBuilder(const builder::LineBuilder& builder)
{
    initGeometry(builder.getMeshDef(), builder.getAttributes());

    if (builder.mRoundCaps)
        mRoundCaps = true;
    else if (builder.mSquareCaps)
        mRoundCaps = false;
    else if (builder.mNoCaps)
        mRoundCaps = false;
    else
        mRoundCaps = true; // default

    if (builder.mNoCaps)
        mNoCaps = true;
    else
        mNoCaps = false; // default

    if (builder.mWorldSpaceSize)
        mWorldSpaceSize = true;
    else if (builder.mScreenSpaceSize)
        mWorldSpaceSize = false;
    else
        mWorldSpaceSize = false; // default

    if (builder.mCameraFacing)
        mCameraFacing = true;
    else if (builder.mOwnNormals)
        mCameraFacing = false;
    else
        mCameraFacing = true; // default

    if (builder.mExtrapolate)
        mExtrapolate = true;
    else
        mExtrapolate = false; // default
}

void LineRenderable::init()
{
    if (mVertexArray)
        return;

    // add missing attributes
    if (!hasAttribute("aLineWidth"))
        addAttribute(detail::make_mesh_attribute("aLineWidth", 5.0f));
    if (!hasAttribute("aNormal"))
        addAttribute(detail::make_mesh_attribute("aNormal", glm::vec3(0, 1, 0)));

    if (getTexturing())
        addAttribute(getTexturing()->mCoordsAttribute);
    else if (getColorMapping())
        addAttribute(getColorMapping()->mDataAttribute);
    else if (!hasAttribute("aColor"))
        addAttribute(detail::make_mesh_attribute("aColor", glm::vec3(mCameraFacing ? 0.0f : 0.25f)));

    const auto aColor = getAttribute("aColor");

    // build meshes
    {
        std::vector<SharedArrayBuffer> abs;

        for (auto const& attr : getAttributes())
        {
            auto ab = attr->createLineRenderableArrayBuffer(*getMeshDefinition());
            if (!ab)
                continue;

            abs.push_back(ab);
        }

        mVertexArray = VertexArray::create(abs, nullptr, GL_LINES);
    }

    // build shader
    {
        detail::MeshShaderBuilder sb;

        sb.addUniform("mat4", "uModel");
        sb.addUniform("mat4", "uInvModel");
        sb.addUniform("bool", "uCameraFacing");
        sb.addUniform("bool", "uRoundCaps");
        sb.addUniform("bool", "uNoCaps");
        sb.addUniform("bool", "uExtrapolate");
        sb.addUniform("bool", "uWorldSpaceSize");
        sb.addUniform("vec2", "uScreenSize");

        sb.addGeometryShaderDecl("#include <glow-viewer/pipelineVertex.glsl>");

        sb.addFragmentShaderDecl("#glow pipeline opaque");
        sb.addFragmentShaderDecl("#include <glow-viewer/pipelineFragment.glsl>");

        sb.addPassthrough("vec3", "Position");
        sb.addPassthrough("vec3", "Normal");
        sb.addPassthrough("float", "LineWidth");

        for (auto const& attr : getAttributes())
            attr->buildShader(sb);

        // colored mesh
        if (aColor)
            sb.addPassthrough(aColor->typeInShader(), "Color");

        // data mapped mesh
        if (getColorMapping())
            getColorMapping()->buildShader(sb);

        // texture mesh
        if (getTexturing())
            getTexturing()->buildShader(sb);

        sb.addGeometryShaderDecl("layout(lines) in;");
        sb.addGeometryShaderDecl("layout(triangle_strip, max_vertices = 8) out;"); // 8 vertices for 6 triangles

        sb.addGeometryShaderDecl("out vec2 gAlpha;");
        sb.addFragmentShaderDecl("in vec2 gAlpha;");

        // Helper functions to emit vertices from the given relative line coordinates
        sb.addGeometryShaderDecl(R"(
void createVertex(vec3 basePos, float ea, vec3 n, vec3 r, vec3 u, float relX, float relY) {
    gAlpha = vec2(relX, relY); // Relative line position for fragment shader
    if(uExtrapolate) {
        if(relX < 0) passthroughMix(0, 1, relX * ea);
        if(relX > 1) passthroughMix(0, 1, (relX - 1) * ea + 1);
    }
    if(relX > 0) relX -= 1; // This sets the factor for the r vector to the right value in case of the second half
    outputViewerVertex(basePos + relX * r + relY * u, uModel);
    vOut.Normal = n; // Set normal again as it might have been overwritten by passthrough
    EmitVertex();
}
void createVertexInverse(vec4 basePos, float ea, vec3 n, vec4 r, vec4 u, float relX, float relY) {
    gAlpha = vec2(relX, relY);
    if(uExtrapolate) {
        if(relX < 0) passthroughMix(0, 1, relX * ea);
        if(relX > 1) passthroughMix(0, 1, (relX - 1) * ea + 1);
    }
    if(relX > 0) relX -= 1;
    outputViewerVertexInverse(basePos + relX * r + relY * u);
    vOut.Normal = n;
    EmitVertex();
}
)");

        sb.addGeometryShaderCode(R"(
    passthrough(0); // Load passthrough data for first vertex

    if(uWorldSpaceSize) {
        vec3 diff = vIn[1].Position - vIn[0].Position;
        vec3 right = normalize(diff);
        vec3 up = uCameraFacing ? normalize(cross(normalize(uViewerGlobals.camPos - vIn[0].Position), right)) : normalize(cross(vIn[0].Normal, right));
        vec3 r = vIn[0].LineWidth * 0.5 * right;
        vec3 u = vIn[0].LineWidth * 0.5 * up;
        vec3 n = uCameraFacing ? vec3(0,0,0) : normalize(mat3(uModel) * vIn[0].Normal); // If cameraFacing: Normal 0 signals unlit rendering
        float ea = vIn[0].LineWidth * 0.5 / length(diff); // Extrapolation alpha

        if(!uNoCaps) {
            createVertex(vIn[0].Position, ea, n, r, u, -1, 1);
            createVertex(vIn[0].Position, ea, n, r, u, -1, -1);
        }
        passthrough(0);
        createVertex(vIn[0].Position, ea, n, r, u, 0, 1);
        createVertex(vIn[0].Position, ea, n, r, u, 0, -1);

        up = uCameraFacing ? normalize(cross(normalize(uViewerGlobals.camPos - vIn[1].Position), right)) : normalize(cross(vIn[1].Normal, right));
        r = vIn[1].LineWidth * 0.5 * right;
        u = vIn[1].LineWidth * 0.5 * up;

        passthrough(1); // Load passthrough data for second vertex
        n = uCameraFacing ? vec3(0,0,0) : normalize(mat3(uModel) * vIn[1].Normal);
        ea = vIn[1].LineWidth * 0.5 / length(diff);

        createVertex(vIn[1].Position, ea, n, r, u, 1, 1);
        createVertex(vIn[1].Position, ea, n, r, u, 1, -1);
        if(!uNoCaps) {
            createVertex(vIn[1].Position, ea, n, r, u, 2, 1);
            createVertex(vIn[1].Position, ea, n, r, u, 2, -1);
        }
    } else {
        // it only makes sense if uCameraFacing is true for screenSpaceSize lines
        vec3 n = vec3(0,0,0);
        vec4 spos0 = uViewerGlobals.proj * uViewerGlobals.view * uModel * vec4(vIn[0].Position, 1.0);
        vec4 spos1 = uViewerGlobals.proj * uViewerGlobals.view * uModel * vec4(vIn[1].Position, 1.0);

        if(spos0.w < 0 || spos1.w < 0) return; // Fix for lines nearer than the near plane

        spos0 /= spos0.w;
        spos1 /= spos1.w;

        vec2 diff = spos1.xy - spos0.xy;
        vec2 right = normalize(diff);
        vec2 up = vec2(-right.y, right.x);
        vec4 r = vec4(vIn[0].LineWidth * 0.5 * right / uScreenSize, 0, 0);
        vec4 u = vec4(vIn[0].LineWidth * 0.5 * up / uScreenSize, 0, 0);
        float ea = length(vIn[0].LineWidth * 0.5 / uScreenSize) / length(diff);

        if(!uNoCaps) {
            createVertexInverse(spos0, ea, n, r, u, -1, 1);
            createVertexInverse(spos0, ea, n, r, u, -1, -1);
        }
        passthrough(0);
        createVertexInverse(spos0, ea, n, r, u, 0, 1);
        createVertexInverse(spos0, ea, n, r, u, 0, -1);

        passthrough(1); // Load passthrough data for second vertex
        r = vec4(vIn[1].LineWidth * 0.5 * right / uScreenSize, 0, 0);
        u = vec4(vIn[1].LineWidth * 0.5 * up / uScreenSize, 0, 0);
        ea = length(vIn[1].LineWidth * 0.5 / uScreenSize) / length(diff);

        createVertexInverse(spos1, ea, n, r, u, 1, 1);
        createVertexInverse(spos1, ea, n, r, u, 1, -1);
        if(!uNoCaps) {
            createVertexInverse(spos1, ea, n, r, u, 2, 1);
            createVertexInverse(spos1, ea, n, r, u, 2, -1);
        }
    }
)");

        sb.addFragmentShaderCode(R"(
    if(uRoundCaps) {
        if(gAlpha.x < 0 && distance(gAlpha, vec2(0, 0)) > 1) discard;
        if(gAlpha.x > 1 && distance(gAlpha, vec2(1, 0)) > 1) discard;
    }
    outputViewerGeometry(vec3(vColor), vNormal);
)");

        mForwardShader = sb.createProgram();
    }

    // prepare IBL
    {
        auto shader = mForwardShader->use();
        material::IBL::initShaderGGX(shader);
    }
}
