#include "MeshRenderable.hh"

#include <glow/common/scoped_gl.hh>

#include <glow-extras/material/IBL.hh>

#include "../SubView.hh"
#include "../ViewRenderContext.hh"
#include "../detail/MeshShaderBuilder.hh"

using namespace glow;
using namespace glow::viewer;

aabb MeshRenderable::computeAabb()
{
    return getMeshAABB().transformed(transform());
}

void MeshRenderable::renderShadow(ViewRenderContext const& ctx, UsedProgram& shader)
{
    // TODO
}

void MeshRenderable::renderDepthPre(ViewRenderContext const& ctx)
{
    GLOW_SCOPED(disable, GL_CULL_FACE);

    auto shader = ctx.use(mDepthPreShader);
    shader.setUniform("uView", ctx.render_camera_data->view);
    shader.setUniform("uProj", ctx.render_camera_data->proj);
    shader.setUniform("uModel", transform());
    mMesh->bind().draw();
}

void MeshRenderable::renderForward(ViewRenderContext const& ctx)
{
    GLOW_SCOPED(disable, GL_CULL_FACE);

    auto shader = ctx.use(mForwardShader);
    shader.setUniform("uModel", transform());
    if (getColorMapping())
        getColorMapping()->prepareShader(shader);
    if (getTexturing())
        getTexturing()->prepareShader(shader);
    for (auto const& a : getAttributes())
        a->prepareShader(shader);
    mMesh->bind().draw();
}

void MeshRenderable::renderTransparent(ViewRenderContext const& ctx)
{
    // TODO
}

SharedMeshRenderable MeshRenderable::create(builder::PolygonBuilder const& builder)
{
    auto r = std::make_shared<MeshRenderable>();
    r->initGeometry(builder.getMeshDef(), builder.getAttributes());
    return r;
}

void MeshRenderable::init()
{
    if (mMesh)
        return;

    // add missing attributes
    if (!hasAttribute("aNormal"))
        addAttribute(getMeshDefinition()->computeFaceNormalAttribute());
    if (getTexturing())
        addAttribute(getTexturing()->mCoordsAttribute);
    else if (getColorMapping())
        addAttribute(getColorMapping()->mDataAttribute);
    else if (!hasAttribute("aColor"))
        addAttribute(detail::make_mesh_attribute("aColor", glm::vec3(1, 1, 1)));

    auto aColor = getAttribute("aColor");

    // build meshes
    {
        std::vector<SharedArrayBuffer> abs;
        SharedArrayBuffer posAB;
        SharedElementArrayBuffer eab;

        for (auto const& attr : getAttributes())
        {
            auto ab = attr->createMeshRenderableArrayBuffer(*getMeshDefinition());
            if (!ab)
                continue;

            abs.push_back(ab);
        }

        mMesh = VertexArray::create(abs, getIndexBuffer());
    }

    // build shader
    {
        mDepthPreShader = Program::createFromFiles({"glow-viewer/depthPre.vsh", "glow-viewer/depthPre.fsh"});

        detail::MeshShaderBuilder sb;

        sb.addUniform("mat4", "uModel");

        sb.addFragmentShaderDecl("#glow pipeline opaque");
        sb.addFragmentShaderDecl("#include <glow-viewer/pipelineFragment.glsl>");

        sb.addVertexShaderDecl("#include <glow-viewer/pipelineVertex.glsl>");

        sb.addPassthrough("vec3", "Normal");

        for (auto const& attr : getAttributes())
            attr->buildShader(sb);

        // colored mesh
        if (aColor)
            sb.addPassthrough(aColor->typeInShader(), "Color");

        // data mapped mesh
        if (getColorMapping())
            getColorMapping()->buildShader(sb);

        // texture mesh
        if (getTexturing())
            getTexturing()->buildShader(sb);

        sb.addVertexShaderCode("outputViewerVertex(aPosition, uModel);");

        sb.addFragmentShaderCode("outputViewerGeometry(vec3(vColor), normalize(vNormal));");

        mForwardShader = sb.createProgram();
    }

    // prepare IBL
    {
        auto shader = mForwardShader->use();
        material::IBL::initShaderGGX(shader);
    }
}
