#include "PointRenderable.hh"

#include <glow-extras/material/IBL.hh>

#include "../SubView.hh"
#include "../ViewRenderContext.hh"
#include "../detail/MeshShaderBuilder.hh"

using namespace glow;
using namespace glow::viewer;

aabb PointRenderable::computeAabb()
{
    // TODO: enlarge by point size
    return getMeshAABB().transformed(transform());
}

void PointRenderable::renderShadow(ViewRenderContext const& ctx, UsedProgram& shader)
{
    // nothing for now
}

void PointRenderable::renderDepthPre(ViewRenderContext const& ctx)
{
    // nothing for now
}

void PointRenderable::renderForward(ViewRenderContext const& ctx)
{
    auto shader = ctx.use(mForwardShader);

    shader.setUniform("uModel", transform());
    shader.setUniform("uInvModel", inverse(transform()));
    shader.setUniform("uScreenSize", glm::vec2(ctx.render_camera_data->resolution));
    if (mRenderMode == RenderMode::Sphere)
    {
        shader.setUniform("uViewTranspose", glm::transpose(ctx.render_camera_data->view));
        shader.setUniform("uTanFov2", ctx.render_camera_data->tanHalfFovY);
        shader.setUniform("uScreenHeight", float(ctx.render_camera_data->resolution.y));
    }

    if (getColorMapping())
        getColorMapping()->prepareShader(shader);
    if (getTexturing())
        getTexturing()->prepareShader(shader);
    for (auto const& a : getAttributes())
        a->prepareShader(shader);

    mVertexArray->bind().draw();
}

void PointRenderable::renderTransparent(ViewRenderContext const& ctx)
{
    // nothing for now
}

SharedPointRenderable PointRenderable::create(const builder::PointBuilder& builder)
{
    auto r = std::make_shared<PointRenderable>();
    r->initFromBuilder(builder);
    return r;
}

void PointRenderable::initFromBuilder(const builder::PointBuilder& builder)
{
    initGeometry(builder.getMeshDef(), builder.getAttributes());

    if (builder.mPointMode == builder::PointBuilder::PointMode::Round)
        mRenderMode = RenderMode::Round;
    else if (builder.mPointMode == builder::PointBuilder::PointMode::Square)
        mRenderMode = RenderMode::Square;
    else if (builder.mPointMode == builder::PointBuilder::PointMode::Spheres)
        mRenderMode = RenderMode::Sphere;
    else
        mRenderMode = RenderMode::Sphere; // default;

    if (builder.mWorldSpaceSize)
        mWorldSpaceSize = true;
    else if (builder.mScreenSpaceSize)
        mWorldSpaceSize = false;
    else
        mWorldSpaceSize = false; // default

    if (builder.mCameraFacing)
        mCameraFacing = true;
    else if (builder.mOwnNormals)
        mCameraFacing = false;
    else
        mCameraFacing = true; // default
}

void PointRenderable::init()
{
    if (mVertexArray)
        return;

    // add missing attributes
    if (!hasAttribute("aPointSize"))
        addAttribute(detail::make_mesh_attribute("aPointSize", 10.0f));
    if (!hasAttribute("aNormal"))
        addAttribute(detail::make_mesh_attribute("aNormal", glm::vec3(0, 1, 0)));

    if (getTexturing())
        addAttribute(getTexturing()->mCoordsAttribute);
    else if (getColorMapping())
        addAttribute(getColorMapping()->mDataAttribute);
    else if (!hasAttribute("aColor"))
    {
        auto const color = glm::vec3(mCameraFacing && mRenderMode != RenderMode::Sphere ? 0.f : .7f);
        addAttribute(detail::make_mesh_attribute("aColor", color));
    }

    auto aColor = getAttribute("aColor");

    // build meshes
    {
        std::vector<SharedArrayBuffer> abs;

        for (auto const& attr : getAttributes())
        {
            auto ab = attr->createPointRenderableArrayBuffer(*getMeshDefinition());
            if (!ab)
                continue;

            ab->setDivisor(1);
            abs.push_back(ab);
        }

        {
            std::vector<glm::vec2> quadPos;

            quadPos.emplace_back(0, 0);
            quadPos.emplace_back(1, 1);
            quadPos.emplace_back(0, 1);

            quadPos.emplace_back(0, 0);
            quadPos.emplace_back(1, 0);
            quadPos.emplace_back(1, 1);

            auto qad = ArrayBuffer::create();
            qad->defineAttribute<glm::vec2>("aQuadPos");
            qad->bind().setData(quadPos);
            abs.push_back(qad);
        }

        mVertexArray = VertexArray::create(abs);
    }

    // build shader
    {
        detail::MeshShaderBuilder sb;

        sb.addUniform("mat4", "uModel");
        sb.addUniform("mat4", "uInvModel");
        sb.addUniform("vec2", "uScreenSize");
        if (mRenderMode == RenderMode::Sphere)
        {
            sb.addUniform("mat4", "uViewTranspose");
            sb.addUniform("float", "uTanFov2");
            sb.addUniform("float", "uScreenHeight");
        }

        sb.addFragmentShaderDecl("#glow pipeline opaque");
        sb.addFragmentShaderDecl("#include <glow-viewer/pipelineFragment.glsl>");

        sb.addVertexShaderDecl("#include <glow-viewer/pipelineVertex.glsl>");
        sb.addAttribute("vec2", "aQuadPos");

        sb.addPassthrough("vec2", "QuadPos");
        sb.addPassthrough("vec3", "Normal");

        for (auto const& attr : getAttributes())
            attr->buildShader(sb);

        // colored mesh
        if (aColor)
            sb.addPassthrough(aColor->typeInShader(), "Color");

        // data mapped mesh
        if (getColorMapping())
            getColorMapping()->buildShader(sb);

        // texture mesh
        if (getTexturing())
            getTexturing()->buildShader(sb);

        if (mRenderMode == RenderMode::Sphere)
        {
            sb.addPassthrough("vec3", "SphereCenter");
            sb.addPassthrough("float", "SphereRadius");

            if (mWorldSpaceSize)
            {
                // world space size spheres
                sb.addVertexShaderCode(R"(
                    vOut.SphereRadius = aPointSize;
                    vOut.SphereCenter = aPosition;

                    vec4 vpos = uViewerGlobals.view * uModel * vec4(aPosition, 1.0);
                    vec3 fwd = normalize(vpos.xyz);
                    vec3 left = normalize(cross(vec3(0,1,0), fwd));
                    vec3 up = cross(fwd, left);
                    vec2 qp = aQuadPos * 2 - 1;
                    vpos.xyz -= left * qp.x * aPointSize;
                    vpos.xyz += up * qp.y * aPointSize;
                    vpos.xyz -= fwd * aPointSize;

                    vpos = uInvModel * uViewerGlobals.invView * vpos;

                    vOut.Normal = vec3(0,0,0);
                    outputViewerVertex(vpos.xyz, uModel);
                )");
            }
            else
            {
                // screen space size spheres
                sb.addVertexShaderCode(R"(
                    float l = 2 * distance(aPosition, uViewerGlobals.camPos) * uTanFov2;
                    float s = l * aPointSize / uScreenHeight;
                    vOut.SphereRadius = s;
                    vOut.SphereCenter = aPosition;

                    vec4 vpos = uViewerGlobals.view * uModel * vec4(aPosition, 1.0);
                    vec3 fwd = normalize(vpos.xyz);
                    vec3 left = normalize(cross(vec3(0,1,0), fwd));
                    vec3 up = cross(fwd, left);
                    vec2 qp = aQuadPos * 2 - 1;
                    vpos.xyz -= left * qp.x * s;
                    vpos.xyz += up * qp.y * s;
                    vpos.xyz -= fwd * s;

                    vpos = uInvModel * uViewerGlobals.invView * vpos;

                    vOut.Normal = vec3(0,0,0);
                    outputViewerVertex(vpos.xyz, uModel);
                )");
            }
        }
        else if (mCameraFacing && !mWorldSpaceSize)
        {
            // camera facing screen space size
            sb.addVertexShaderCode(R"(
                vec4 spos = uViewerGlobals.proj * uViewerGlobals.view * uModel * vec4(aPosition, 1.0);
                spos /= spos.w;

                vec2 qp = aQuadPos * 2 - 1;
                spos.xy += qp * aPointSize / uScreenSize;

                spos = uViewerGlobals.invProj * spos;
                spos /= spos.w;
                spos = uInvModel * uViewerGlobals.invView * spos;

                vOut.Normal = vec3(0,0,0);
                outputViewerVertex(spos.xyz, uModel);
            )");
        }
        else if (mCameraFacing)
        {
            // camera facing world space size
            sb.addVertexShaderCode(R"(
                vec4 vpos = uViewerGlobals.view * uModel * vec4(aPosition, 1.0);
                vec2 qp = aQuadPos * 2 - 1;
                vpos.xy += qp * aPointSize;

                vpos = uInvModel * uViewerGlobals.invView * vpos;

                vOut.Normal = vec3(0,0,0);
                outputViewerVertex(vpos.xyz, uModel);
            )");
        }
        else
        {
            // Normal aligned world space size
            sb.addVertexShaderCode(R"(
                vec3 dir = normalize(mat3(uModel) * aNormal);
                vec3 up = abs(dir.y) < abs(dir.x) ? vec3(0, 1, 0) : vec3(1, 0, 0);
                vec3 left = normalize(cross(dir, up));
                up = normalize(cross(dir, left));
                vec2 qp = aQuadPos - 0.5;
                vec3 pos = aPosition + aPointSize * (left * qp.x + up * qp.y);

                vOut.Normal = dir;
                outputViewerVertex(pos, uModel);
            )");
        }

        if (mRenderMode == RenderMode::Sphere)
        {
            sb.addFragmentShaderDecl(R"(
                 float sphere(vec3 ray, vec3 dir, vec3 center, float radius)
                 {
                     vec3 rc = ray-center;
                     float c = dot(rc, rc) - (radius * radius);
                     float b = dot(dir, rc);
                     float d = b*b - c;
                     float t = -b - sqrt(abs(d));
                     float st = step(0.0, min(t,d));
                     return mix(-1.0, t, st);
                 }

                 float distance2(vec3 a, vec3 b)
                 {
                     vec3 d = a - b;
                     return dot(d, d);
                 }
            )");

            sb.addFragmentShaderCode(R"(
                vec3 rayOrigin = uViewerGlobals.camPos;
                vec3 rayDir = normalize(vIn.fragPosWS - uViewerGlobals.camPos);

                vec3 closestP = rayOrigin + rayDir * dot(rayDir, vSphereCenter - rayOrigin);
                float sphereDis2 = distance2(closestP, vSphereCenter);

                if (sphereDis2 > vSphereRadius * vSphereRadius)
                    discard;

                vec3 spherePos = closestP - rayDir * sqrt(vSphereRadius * vSphereRadius - sphereDis2);

                vec3 sphereN = normalize(spherePos - vSphereCenter);

                vec4 spherePosCS = uViewerGlobals.proj * uViewerGlobals.view * vec4(spherePos, 1);
                float depthNDC = spherePosCS.z / spherePosCS.w;

                gl_FragDepth = (depthNDC - gl_DepthRange.near) / gl_DepthRange.diff;

                outputViewerGeometry(vec3(vColor), sphereN);
            )");
        }
        else
        {
            // Quads and Circles
            if (mRenderMode == RenderMode::Round)
                sb.addFragmentShaderCode("if (distance(vQuadPos, vec2(0.5)) > 0.5) discard;");

            sb.addFragmentShaderCode("outputViewerGeometry(vec3(vColor), vNormal);");
        }

        mForwardShader = sb.createProgram();
    }

    // prepare IBL
    {
        auto shader = mForwardShader->use();
        material::IBL::initShaderGGX(shader);
    }
}
