#pragma once

#include <cassert>
#include <utility>

#include <glow/common/nodiscard.hh>
#include <glow/common/non_copyable.hh>

#include "Scene.hh"
#include "SubView.hh"
#include "layouts/layouts.hh"
#include "macros.hh"
#include "renderables/Renderable.hh"

// the two important API headers:
#include "configure.hh"
#include "make_renderable.hh"

// not really needed by often used:
#include <glow-extras/colors/color.hh>
#include "materials/materials.hh"
#include "objects/objects.hh"

/**
 * One-size-fits-most function for visualizing meshes, properties, ...
 *
 * view(...): creates a view and shows it
 *
 * See viewer/Readme.md for advanced usages
 */

namespace glow
{
namespace viewer
{
/// creates an empty 3D view
GLOW_NODISCARD detail::raii_view view();

/// creates a 3D view, adds `obj` as renderable and calls v->configure(renderable, arg) for each arg in args
template <class ObjT, class... Args>
detail::raii_view view(ObjT const& obj, Args const&... args);

/// creates an empty view with column layout
/// Usage:
///   auto c = viewer::columns();
///   c.view(pos, ...);
///   c.view(pos, ...);
///   ...
///   // show on end-of-scope
GLOW_NODISCARD detail::raii_view columns();

/// creates an empty view with row layout
/// Usage:
///   auto r = viewer::rows();
///   r.view(pos, ...);
///   r.view(pos, ...);
///   ...
///   // show on end-of-scope
GLOW_NODISCARD detail::raii_view rows();

/// creates an empty view with grid layout
/// Usage:
///   auto g = viewer::grid(2, 3);
///   g.view(pos, ...);
///   g.view(pos, ...);
///   ...
///   // show on end-of-scope
GLOW_NODISCARD detail::raii_view grid(int cols, int rows);
GLOW_NODISCARD detail::raii_view grid(float aspect = 1.0f); // aspect = w / h

// ======= Implementation ========

class Viewer
{
private:
    static SharedTextureCubeMap sViewerSkybox;

public:
    static void GlobalInit(SharedTextureCubeMap const& skybox = nullptr);

    static SharedTextureCubeMap const& getSkybox() noexcept { return sViewerSkybox; }
};

namespace detail
{
// shows view on scope end
struct raii_view
{
    GLOW_RAII_CLASS(raii_view);

    static std::vector<SharedSubView> sSubViewStack;

public:
    raii_view(SharedSubView const& view) : mView(view) { sSubViewStack.push_back(mView); }
    raii_view(raii_view&& rhs) noexcept : mView(std::move(rhs.mView)) {}
    ~raii_view()
    {
        if (mView)
        {
            assert(!sSubViewStack.empty() && "corrupted subview stack");
            assert(sSubViewStack.back() == mView && "corrupted subview stack");
            sSubViewStack.pop_back();

            if (sSubViewStack.empty()) // last one? show!
                mView->show();
            else // otherwise, attach
                sSubViewStack.back()->addSubView(mView);
        }
    }

    SubView* operator->() const { return mView.get(); }

    raii_view view() const { return ::glow::viewer::view(); }
    template <class ObjT, class... Args>
    raii_view view(ObjT const& obj, Args const&... args) const
    {
        return ::glow::viewer::view(obj, args...);
    }

    detail::raii_view columns() { return ::glow::viewer::columns(); }
    detail::raii_view rows() { return ::glow::viewer::rows(); }
    detail::raii_view grid(int cols, int rows) { return ::glow::viewer::grid(cols, rows); }

private:
    SharedSubView mView;
};

inline SharedSubView getOrCreateView()
{
    // if current view is not container, it will be used to place renderables in it
    if (!detail::raii_view::sSubViewStack.empty() && !detail::raii_view::sSubViewStack.back()->isContainer())
        return detail::raii_view::sSubViewStack.back();

    // otherwise a new subview is created
    return SubView::create();
}
}

inline detail::raii_view view()
{
    auto v = detail::getOrCreateView();
    return v;
}

template <class ObjT, class... Args>
detail::raii_view view(ObjT const& obj, Args const&... args)
{
    auto v = detail::getOrCreateView();
    auto s = v->scene();
    auto r = make_renderable(obj);
    s->add(r);

    // A compile error here means configure(v, s, r, arg) was not found for one of your args
    // see configure.hh
    int a[] = {0, (configure(v, s, *r, args), 0)...};
    (void)a; // suppress unused warning

    return v;
}

inline detail::raii_view columns() { return grid(-1, 1); }

inline detail::raii_view rows() { return grid(1, -1); }

inline detail::raii_view grid(int cols, int rows)
{
    auto v = detail::getOrCreateView();
    v->setContainer(true);
    v->setLayout(std::make_shared<GridLayout>(cols, rows));
    return v;
}

inline detail::raii_view grid(float aspect)
{
    auto v = detail::getOrCreateView();
    v->setContainer(true);
    v->setLayout(std::make_shared<GridLayout>(-1, -1, aspect));
    return v;
}
}
}

// insert into namespaces so ADL works
// e.g. view(pos) works without using any namespace (pos is polymesh::vertex_attribute<glm::vec3>)
GLOW_VIEWER_INJECT_FUNCTION(glow::viewer::view);
